@@ -35,6 +35,7 @@ from hy.models.float import HyFloat
 from hy.models.list import HyList
 from hy.models.dict import HyDict
 
+import hy.importer
 from hy.core import process
 
 from hy.util import str_type
@@ -93,6 +94,14 @@ def ast_str(foobar):
 
 
 def builds(_type):
+
+    unpythonic_chars = ["-"]
+    really_ok = ["-"]
+    if True in (x in str_type(_type) for x in unpythonic_chars):
+        if _type not in really_ok:
+            raise TypeError("`build' needs to be *post* translated strings, "
+                            "Mr. / Mrs. Hypser. -- `%s' sucks." % (_type))
+
     def _dec(fn):
         _compile_table[_type] = fn
         return fn
@@ -169,33 +178,37 @@ class Result(object):
 
         If there is no expression context, we return a "None" expression.
         """
-        if not self.expr:
-            # Spoof the position of the last statement for our generated None
-            lineno = 0
-            col_offset = 0
-            if self.stmts:
-                lineno = self.stmts[-1].lineno
-                col_offset = self.stmts[-1].col_offset
-
-            return ast.Name(id=ast_str("None"),
-                            arg=ast_str("None"),
-                            ctx=ast.Load(),
-                            lineno=lineno,
-                            col_offset=col_offset)
-            # XXX: Likely raise Exception here - this will assertionfail
-            #      pypy since the ast will be out of numerical order.
-        else:
+        if self.expr:
             return self.expr
 
+        # Spoof the position of the last statement for our generated None
+        lineno = 0
+        col_offset = 0
+        if self.stmts:
+            lineno = self.stmts[-1].lineno
+            col_offset = self.stmts[-1].col_offset
+
+        return ast.Name(id=ast_str("None"),
+                        arg=ast_str("None"),
+                        ctx=ast.Load(),
+                        lineno=lineno,
+                        col_offset=col_offset)
+        # XXX: Likely raise Exception here - this will assertionfail
+        #      pypy since the ast will be out of numerical order.
+
     def expr_as_stmt(self):
         """Convert the Result's expression context to a statement
 
         This is useful when we want to use the stored expression in a
         statement context (for instance in a code branch).
 
+        We drop ast.Names if they are appended to statements, as they
+        can't have any side effect. "Bare" names still get converted to
+        statements.
+
         If there is no expression context, return an empty result.
         """
-        if self.expr:
+        if self.expr and not (isinstance(self.expr, ast.Name) and self.stmts):
             return Result() + ast.Expr(lineno=self.expr.lineno,
                                        col_offset=self.expr.col_offset,
                                        value=self.expr)
@@ -257,20 +270,6 @@ class Result(object):
         )
 
 
-def _collect(results):
-    """Collect the expression contexts from a list of results
-
-    This returns a list of the expression contexts, and the sum of the Result
-    objects passed as arguments.
-    """
-    compiled_exprs = []
-    ret = Result()
-    for result in results:
-        ret += result
-        compiled_exprs.append(ret.force_expr)
-    return compiled_exprs, ret
-
-
 def _branch(results):
     """Make a branch out of a list of Result objects
 
@@ -344,15 +343,24 @@ class HyASTCompiler(object):
         """Convert the Result's imports to statements"""
         ret = Result()
         for module, names in self.imports.items():
-            ret += self.compile([
-                HyExpression([
-                    HySymbol("import"),
-                    HyList([
+            if None in names:
+                ret += self.compile([
+                    HyExpression([
+                        HySymbol("import"),
                         HySymbol(module),
-                        HyList([HySymbol(name) for name in sorted(names)])
-                    ])
-                ]).replace(expr)
-            ])
+                    ]).replace(expr)
+                ])
+            names = sorted(name for name in names if name)
+            if names:
+                ret += self.compile([
+                    HyExpression([
+                        HySymbol("import"),
+                        HyList([
+                            HySymbol(module),
+                            HyList([HySymbol(name) for name in names])
+                        ])
+                    ]).replace(expr)
+                ])
         self.imports = defaultdict(set)
         return ret.stmts
 
@@ -363,8 +371,6 @@ class HyASTCompiler(object):
             if not isinstance(ret, Result):
                 ret = Result() + ret
             return ret
-        else:
-            return None
 
     def compile(self, tree):
         try:
@@ -384,7 +390,18 @@ class HyASTCompiler(object):
         raise HyCompileError(Exception("Unknown type: `%s'" % _type))
 
     def _compile_collect(self, exprs):
-        return _collect(self.compile(expr) for expr in exprs)
+        """Collect the expression contexts from a list of compiled expression.
+
+        This returns a list of the expression contexts, and the sum of the
+        Result objects passed as arguments.
+
+        """
+        compiled_exprs = []
+        ret = Result()
+        for expr in exprs:
+            ret += self.compile(expr)
+            compiled_exprs.append(ret.force_expr)
+        return compiled_exprs, ret
 
     def _compile_branch(self, exprs):
         return _branch(self.compile(expr) for expr in exprs)
@@ -407,6 +424,9 @@ class HyASTCompiler(object):
                 if expr == "&rest" and lambda_keyword is None:
                     lambda_keyword = expr
                 elif expr == "&optional":
+                    if len(defaults) > 0:
+                        raise HyCompileError("There can only be &optional "
+                                             "arguments or one &key argument")
                     lambda_keyword = expr
                 elif expr == "&key":
                     lambda_keyword = expr
@@ -432,8 +452,8 @@ class HyASTCompiler(object):
                                     "argument")
                 else:
                     if len(defaults) > 0:
-                        raise HyCompileError("There can only be "
-                                             "one &key argument")
+                        raise HyCompileError("There can only be &optional "
+                                             "arguments or one &key argument")
                     # As you can see, Python has a funny way of
                     # defining keyword arguments.
                     for k, v in expr.items():
@@ -441,8 +461,17 @@ class HyASTCompiler(object):
                         ret += self.compile(v)
                         defaults.append(ret.force_expr)
             elif lambda_keyword == "&optional":
-                # not implemented yet.
-                pass
+                if isinstance(expr, HyList):
+                    if not len(expr) == 2:
+                        raise TypeError("optional args should be bare names "
+                                        "or 2-item lists")
+                    k, v = expr
+                else:
+                    k = expr
+                    v = HySymbol("None").replace(k)
+                args.append(k)
+                ret += self.compile(v)
+                defaults.append(ret.force_expr)
             elif lambda_keyword == "&kwargs":
                 if kwargs:
                     raise HyCompileError("There can only be one "
@@ -483,33 +512,86 @@ class HyASTCompiler(object):
         ret += ret.expr_as_stmt()
         return ret
 
-    def _render_quoted_form(self, form):
+    def _render_quoted_form(self, form, level):
+        """
+        Render a quoted form as a new HyExpression.
+
+        `level` is the level of quasiquoting of the current form. We can
+        unquote if level is 0.
+
+        Returns a three-tuple (`imports`, `expression`, `splice`).
+
+        The `splice` return value is used to mark `unquote-splice`d forms.
+        We need to distinguish them as want to concatenate them instead of
+        just nesting them.
+        """
+        if level == 0:
+            if isinstance(form, HyExpression):
+                if form and form[0] in ("unquote", "unquote_splice"):
+                    if len(form) != 2:
+                        raise HyTypeError(form,
+                                          ("`%s' needs 1 argument, got %s" %
+                                           form[0], len(form) - 1))
+                    return set(), form[1], (form[0] == "unquote_splice")
+
+        if isinstance(form, HyExpression):
+            if form and form[0] == "quasiquote":
+                level += 1
+            if form and form[0] in ("unquote", "unquote_splice"):
+                level -= 1
+
         name = form.__class__.__name__
-        imports = [name]
+        imports = set([name])
 
         if isinstance(form, HyList):
-            contents = []
+            if not form:
+                contents = HyList()
+            else:
+                # If there are arguments, they can be spliced
+                # so we build a sum...
+                contents = HyExpression([HySymbol("+"), HyList()])
+
             for x in form:
-                form_imports, form_contents = self._render_quoted_form(x)
-                imports += form_imports
-                contents.append(form_contents)
-            return imports, HyExpression(
-                [HySymbol(name),
-                 HyList(contents)]
-            ).replace(form)
+                f_imports, f_contents, splice = self._render_quoted_form(x,
+                                                                         level)
+                imports.update(f_imports)
+                if splice:
+                    to_add = f_contents
+                else:
+                    to_add = HyList([f_contents])
+
+                contents.append(to_add)
+
+            return imports, HyExpression([HySymbol(name),
+                                          contents]).replace(form), False
+
         elif isinstance(form, HySymbol):
             return imports, HyExpression([HySymbol(name),
-                                          HyString(form)]).replace(form)
-        return imports, HyExpression([HySymbol(name), form]).replace(form)
+                                          HyString(form)]).replace(form), False
+
+        return imports, HyExpression([HySymbol(name),
+                                      form]).replace(form), False
 
     @builds("quote")
+    @builds("quasiquote")
     @checkargs(exact=1)
     def compile_quote(self, entries):
-        imports, stmts = self._render_quoted_form(entries[1])
+        if entries[0] == "quote":
+            # Never allow unquoting
+            level = float("inf")
+        else:
+            level = 0
+        imports, stmts, splice = self._render_quoted_form(entries[1], level)
         ret = self.compile(stmts)
         ret.add_imports("hy", imports)
         return ret
 
+    @builds("unquote")
+    @builds("unquote_splicing")
+    def compile_unquote(self, expr):
+        raise HyTypeError(expr,
+                          "`%s' can't be used at the top-level" % expr[0])
+
     @builds("eval")
     @checkargs(exact=1)
     def compile_eval(self, expr):
@@ -562,8 +644,24 @@ class HyASTCompiler(object):
         # (try something…)
         body = self.compile(body)
 
-        # XXX we will likely want to make this a tempvar
-        body += body.expr_as_stmt()
+        var = self.get_anon_var()
+        name = ast.Name(id=ast_str(var), arg=ast_str(var),
+                        ctx=ast.Store(),
+                        lineno=expr.start_line,
+                        col_offset=expr.start_column)
+
+        expr_name = ast.Name(id=ast_str(var), arg=ast_str(var),
+                             ctx=ast.Load(),
+                             lineno=expr.start_line,
+                             col_offset=expr.start_column)
+
+        returnable = Result(expr=expr_name, temp_variables=[expr_name, name])
+
+        body += ast.Assign(targets=[name],
+                           value=body.force_expr,
+                           lineno=expr.start_line,
+                           col_offset=expr.start_column)
+
         body = body.stmts
         if not body:
             body = [ast.Pass(lineno=expr.start_line,
@@ -579,7 +677,7 @@ class HyASTCompiler(object):
                 raise HyTypeError(e, "Empty list not allowed in `try'")
 
             if e[0] in (HySymbol("except"), HySymbol("catch")):
-                handler_results += self.compile(e)
+                handler_results += self._compile_catch_expression(e, name)
                 handlers.append(handler_results.stmts.pop())
             elif e[0] == HySymbol("else"):
                 if orelse:
@@ -631,7 +729,7 @@ class HyASTCompiler(object):
                 body=body,
                 handlers=handlers,
                 orelse=orelse,
-                finalbody=finalbody)
+                finalbody=finalbody) + returnable
 
         if finalbody:
             if handlers:
@@ -644,30 +742,34 @@ class HyASTCompiler(object):
                         handlers=handlers,
                         body=body,
                         orelse=orelse)],
-                    finalbody=finalbody)
+                    finalbody=finalbody) + returnable
 
             return ret + ast.TryFinally(
                 lineno=expr.start_line,
                 col_offset=expr.start_column,
                 body=body,
-                finalbody=finalbody)
+                finalbody=finalbody) + returnable
 
         return ret + ast.TryExcept(
             lineno=expr.start_line,
             col_offset=expr.start_column,
             handlers=handlers,
             body=body,
-            orelse=orelse)
+            orelse=orelse) + returnable
 
-    @builds("catch")
     @builds("except")
-    def compile_catch_expression(self, expr):
+    @builds("catch")
+    def magic_internal_form(self, expr):
+        raise TypeError("Error: `%s' can't be used like that." % (expr[0]))
+
+    def _compile_catch_expression(self, expr, var):
         catch = expr.pop(0)  # catch
 
         try:
             exceptions = expr.pop(0)
         except IndexError:
             exceptions = HyList()
+
         # exceptions catch should be either:
         # [[list of exceptions]]
         # or
@@ -678,6 +780,7 @@ class HyASTCompiler(object):
         # [exception]
         # or
         # []
+
         if not isinstance(exceptions, HyList):
             raise HyTypeError(exceptions,
                               "`%s' exceptions list is not a list" % catch)
@@ -725,7 +828,10 @@ class HyASTCompiler(object):
                               "`%s' needs a valid exception list" % catch)
 
         body = self._compile_branch(expr)
-        # XXX tempvar handling magic
+        body += ast.Assign(targets=[var],
+                           value=body.force_expr,
+                           lineno=expr.start_line,
+                           col_offset=expr.start_column)
         body += body.expr_as_stmt()
 
         body = body.stmts
@@ -846,31 +952,6 @@ class HyASTCompiler(object):
                           lineno=e.start_line,
                           col_offset=e.start_column)
 
-    @builds("lambda")
-    @checkargs(2)
-    def compile_lambda_expression(self, expr):
-        expr.pop(0)
-        sig = expr.pop(0)
-        body = self.compile(expr.pop(0))
-
-        body += ast.Lambda(
-            lineno=expr.start_line,
-            col_offset=expr.start_column,
-            args=ast.arguments(args=[
-                ast.Name(arg=ast_str(x), id=ast_str(x),
-                         ctx=ast.Param(),
-                         lineno=x.start_line,
-                         col_offset=x.start_column)
-                for x in sig],
-                vararg=None,
-                kwarg=None,
-                defaults=[],
-                kwonlyargs=[],
-                kw_defaults=[]),
-            body=body.force_expr)
-
-        return body
-
     @builds("yield")
     @checkargs(max=1)
     def compile_yield_expression(self, expr):
@@ -1110,24 +1191,14 @@ class HyASTCompiler(object):
     def compile_kwapply_expression(self, expr):
         expr.pop(0)  # kwapply
         call = self.compile(expr.pop(0))
-        kwargs = expr.pop(0)
+        kwargs = self.compile(expr.pop(0))
 
         if type(call.expr) != ast.Call:
             raise HyTypeError(expr, "kwapplying a non-call")
 
-        if type(kwargs) != HyDict:
-            raise TypeError("kwapplying with a non-dict")
+        call.expr.kwargs = kwargs.force_expr
 
-        keywords = []
-        ret = Result()
-        for x in kwargs:
-            ret += self.compile(kwargs[x])
-            keywords.append(ast.keyword(arg=ast_str(x),
-                                        value=ret.force_expr))
-
-        call.expr.keywords = keywords
-
-        return ret + call
+        return kwargs + call
 
     @builds("not")
     @builds("~")
@@ -1144,6 +1215,18 @@ class HyASTCompiler(object):
                                col_offset=operator.start_column)
         return operand
 
+    @builds("require")
+    def compile_require(self, expression):
+        """
+        TODO: keep track of what we've imported in this run and then
+        "unimport" it after we've completed `thing' so that we don't polute
+        other envs.
+        """
+        expression.pop(0)
+        for entry in expression:
+            __import__(entry)  # Import it fo' them macros.
+        return Result()
+
     @builds("and")
     @builds("or")
     @checkargs(min=2)
@@ -1331,9 +1414,13 @@ class HyASTCompiler(object):
     @builds("setv")
     @checkargs(2)
     def compile_def_expression(self, expression):
-        expression.pop(0)
-        name = expression.pop(0)
-        result = self.compile(expression.pop(0))
+        return self._compile_assign(expression[1], expression[2],
+                                    expression.start_line,
+                                    expression.start_column)
+
+    def _compile_assign(self, name, result,
+                        start_line, start_column):
+        result = self.compile(result)
 
         if result.temp_variables and isinstance(name, HyString):
             result.rename(name)
@@ -1343,8 +1430,8 @@ class HyASTCompiler(object):
         st_name = self._storeize(ld_name)
 
         result += ast.Assign(
-            lineno=expression.start_line,
-            col_offset=expression.start_column,
+            lineno=start_line,
+            col_offset=start_column,
             targets=[st_name], value=result.force_expr)
 
         result += ld_name
@@ -1411,16 +1498,37 @@ class HyASTCompiler(object):
                         col_offset=expression.start_column)
         return ret
 
+    @builds("lambda")
     @builds("fn")
     @checkargs(min=1)
     def compile_function_def(self, expression):
-        expression.pop(0)
-
-        name = self.get_anon_fn()
+        called_as = expression.pop(0)
 
         arglist = expression.pop(0)
         ret, args, defaults, stararg, kwargs = self._parse_lambda_list(arglist)
+
+        args = ast.arguments(
+            args=[ast.Name(arg=ast_str(x), id=ast_str(x),
+                           ctx=ast.Param(),
+                           lineno=x.start_line,
+                           col_offset=x.start_column)
+                  for x in args],
+            vararg=stararg,
+            kwarg=kwargs,
+            kwonlyargs=[],
+            kw_defaults=[],
+            defaults=defaults)
+
         body = self._compile_branch(expression)
+        if not body.stmts and called_as == "lambda":
+            ret += ast.Lambda(
+                lineno=expression.start_line,
+                col_offset=expression.start_column,
+                args=args,
+                body=body.force_expr)
+
+            return ret
+
         if body.expr:
             body += ast.Return(value=body.expr,
                                lineno=body.expr.lineno,
@@ -1430,22 +1538,12 @@ class HyASTCompiler(object):
             body += ast.Pass(lineno=expression.start_line,
                              col_offset=expression.start_column)
 
+        name = self.get_anon_fn()
+
         ret += ast.FunctionDef(name=name,
                                lineno=expression.start_line,
                                col_offset=expression.start_column,
-                               args=ast.arguments(
-                                   args=[
-                                       ast.Name(
-                                           arg=ast_str(x), id=ast_str(x),
-                                           ctx=ast.Param(),
-                                           lineno=x.start_line,
-                                           col_offset=x.start_column)
-                                       for x in args],
-                                   vararg=stararg,
-                                   kwarg=kwargs,
-                                   kwonlyargs=[],
-                                   kw_defaults=[],
-                                   defaults=defaults),
+                               args=args,
                                body=body.stmts,
                                decorator_list=[])
 
@@ -1459,6 +1557,80 @@ class HyASTCompiler(object):
 
         return ret
 
+    @builds("defclass")
+    @checkargs(min=1)
+    def compile_class_expression(self, expression):
+        expression.pop(0)  # class
+
+        class_name = expression.pop(0)
+
+        if expression:
+            base_list = expression.pop(0)
+            if not isinstance(base_list, HyList):
+                raise HyTypeError(expression,
+                                  "Bases class must be a list")
+            bases_expr, bases = self._compile_collect(base_list)
+        else:
+            bases_expr = []
+            bases = Result()
+
+        body = Result()
+
+        if expression:
+            try:
+                body_expression = iter(expression.pop(0))
+            except TypeError:
+                raise HyTypeError(
+                    expression,
+                    "Wrong argument type for defclass slots definition.")
+            for b in body_expression:
+                if len(b) != 2:
+                    raise HyTypeError(
+                        expression,
+                        "Wrong number of argument in defclass slot.")
+                body += self._compile_assign(b[0], b[1],
+                                             b.start_line, b.start_column)
+                body += body.expr_as_stmt()
+
+        if not body.stmts:
+            body += ast.Pass(lineno=expression.start_line,
+                             col_offset=expression.start_column)
+
+        return bases + ast.ClassDef(
+            lineno=expression.start_line,
+            col_offset=expression.start_column,
+            decorator_list=[],
+            name=ast_str(class_name),
+            keywords=[],
+            starargs=None,
+            kwargs=None,
+            bases=bases_expr,
+            body=body.stmts)
+
+    @builds("defmacro")
+    @checkargs(min=1)
+    def compile_macro(self, expression):
+        expression.pop(0)
+        name = expression.pop(0)
+        if not isinstance(name, HySymbol):
+            raise HyTypeError(name, ("received a `%s' instead of a symbol "
+                                     "for macro name" % type(name).__name__))
+        name = HyString(name).replace(name)
+        new_expression = HyExpression([
+            HySymbol("with_decorator"),
+            HyExpression([HySymbol("hy.macros.macro"), name]),
+            HyExpression([HySymbol("fn")] + expression),
+        ]).replace(expression)
+
+        # Compile-time hack: we want to get our new macro now
+        hy.importer.hy_eval(new_expression, {'hy': hy})
+
+        # We really want to have a `hy` import to get hy.macro in
+        ret = self.compile(new_expression)
+        ret.add_imports('hy', [None])
+
+        return ret
+
     @builds(HyInteger)
     def compile_integer(self, number):
         return ast.Num(n=int(number),
