@@ -36,13 +36,15 @@ def macro(name):
 def process(tree):
     if isinstance(tree, HyExpression):
         fn = tree[0]
+        if fn in ("quote", "quasiquote"):
+            return tree
         ntree = HyExpression([fn] + [process(x) for x in tree[1:]])
         ntree.replace(tree)
 
         if isinstance(fn, HyString):
             if fn in _hy_macros:
                 m = _hy_macros[fn]
-                obj = m(ntree)
+                obj = m(*ntree[1:])
                 obj.replace(tree)
                 return obj
 
